-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore BENCH.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- Test bench for MicroCore

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;
                                
ENTITY bench IS
END bench;

ARCHITECTURE testbench OF bench IS

COMPONENT core
PORT (reset       : IN    STD_LOGIC;
      done        : IN    STD_LOGIC;
      sysclk      : IN    STD_LOGIC; -- at least twice the uClk frequency
      clk_en      : IN    STD_LOGIC;
      except      : IN    STD_LOGIC;
      int_in      : IN    int_bus;
      clk         : OUT   STD_LOGIC; -- reference clock for peripheral I/O
      sel_ram     : OUT   STD_LOGIC;
      sel_io      : OUT   STD_LOGIC;
      ioWr        : OUT   STD_LOGIC;
      iowrite     : OUT   STD_LOGIC;
      ioaddr      : OUT   data_addr;
      iodata      : INOUT data_bus;
      romWr       : OUT   STD_LOGIC;
      romwrite    : OUT   STD_LOGIC;
      romaddr     : INOUT prog_addr;
      romdata     : INOUT inst_bus);
END COMPONENT;

COMPONENT program_memory
PORT (clk   : IN    STD_LOGIC;
      we    : IN    STD_LOGIC;
      addr  : IN    prog_addr;
      di    : IN    inst_bus;
      do    : OUT   inst_bus);
END COMPONENT;

SIGNAL done       : STD_LOGIC;
SIGNAL reset      : STD_LOGIC; 
SIGNAL sysclk     : STD_LOGIC;
SIGNAL clk        : STD_LOGIC;
SIGNAL clk_en     : STD_LOGIC;
SIGNAL except     : STD_LOGIC;
SIGNAL interrupt  : int_bus;

SIGNAL sel_io     : STD_LOGIC;
SIGNAL sel_ram    : STD_LOGIC;
SIGNAL ioWr       : STD_LOGIC;
SIGNAL iowrite    : STD_LOGIC;
SIGNAL ioaddr     : data_addr;
SIGNAL iodata     : data_bus;
SIGNAL ioreg      : data_bus; 
SIGNAL ioreg2     : data_bus;

SIGNAL ramwrite   : STD_LOGIC;
SIGNAL ramaddr    : data_addr;
SIGNAL ramdata    : data_bus;
SIGNAL romdata    : inst_bus;
SIGNAL romdin     : inst_bus;
SIGNAL romaddr    : prog_addr;
SIGNAL romWr      : STD_LOGIC;
SIGNAL romwrite   : STD_LOGIC;

CONSTANT IOPORT1  : data_addr := to_vec(-16, data_addr_width); -- for testbench
CONSTANT IOPORT2  : data_addr := to_vec(-17, data_addr_width); -- for testbench

CONSTANT cycle    : TIME := 100 ns;

BEGIN

fpgacore: core PORT 
MAP ( reset       => reset,
      done        => done,
      sysclk      => sysclk,
      clk_en      => clk_en,
      except      => except,
      int_in      => interrupt,
      clk         => clk,
      sel_ram     => sel_ram,
      sel_io      => sel_io,
      ioWr        => ioWr,
      iowrite     => iowrite,
      ioaddr      => ioaddr,
      iodata      => iodata,
      romWr       => romWr,
      romwrite    => romwrite,
      romaddr     => romaddr,
      romdata     => romdata);

iodata <= (OTHERS => 'Z')  WHEN  iowrite='1'  ELSE
          ramdata          WHEN  sel_ram='1'  ELSE
          ioreg            WHEN  sel_io='1'  AND ioaddr=IOPORT1  ELSE
          ioreg2           WHEN  sel_io='1'  AND ioaddr=IOPORT2  ELSE
          to_vec(85, data_width);

data_memory: external_ram GENERIC MAP(data_width, data_addr_width)
PORT MAP(enable => sel_ram,
         write  => ioWr,
         addr   => ioaddr,
         data   => ramdata);
ramdata <= iodata WHEN  iowrite='1'  ELSE (OTHERS => 'Z');

prog_mem: program_memory
PORT MAP(clk  => sysclk,
         we   => romWr,
         addr => romaddr,
         di   => romdata,
         do   => romdin);
romdata <= romdin WHEN  romwrite='0'  ELSE (OTHERS => 'Z');

reset_core: PROCESS
BEGIN
   done         <= '0';
   reset        <= '1';
   clk_en       <= '1';
   interrupt    <= "00";
   except       <= '0';
   WAIT FOR cycle+20 ns;
   done         <= '1';
   WAIT FOR cycle+20 ns;
   reset        <= '0';
   WAIT;
END PROCESS reset_core;

testio: PROCESS(reset, sysclk)
BEGIN
  IF  reset='1'  THEN
     ioreg <= (OTHERS => '0');
     ioreg2 <= (OTHERS => '0');
  ELSIF  rising_edge(sysclk)  THEN
     IF  clk_en='1' AND sel_io='1' AND iowrite='1'  THEN
       IF  ioaddr=IOPORT1  THEN
          ioreg <= iodata;
       END IF;
       IF  ioaddr=IOPORT2  THEN
          ioreg2 <= iodata;
       END IF;
     END IF;
  END IF;
END PROCESS testio;

clock: PROCESS
BEGIN
  LOOP
    sysclk <= '1';
    WAIT FOR cycle/4;
    sysclk <= '0';
    WAIT FOR cycle/4;
  END LOOP;
END PROCESS clock;

END testbench;
