-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore CONSTANTS.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- Defining all MicroCore bus-widths and records, as well as the 
-- opcode fields.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     work.functions.ALL;

PACKAGE constants IS

CONSTANT syn_stackram      : STD_LOGIC := '0'; -- '1' when stack ram will be realised by synchronous block ram
CONSTANT with_locals       : STD_LOGIC := '1'; -- '1' when RSP+TOS addressing mode
CONSTANT with_tasks        : STD_LOGIC := '1'; -- '1' when TASK+TOS addressing mode
CONSTANT with_nos          : STD_LOGIC := '1'; -- '1' when NOS register will be created for single cycle SWAP
CONSTANT with_tor          : STD_LOGIC := '1'; -- '1' when TOR register will be created for MULTS, UDIVS, and NEXT
CONSTANT with_ip           : STD_LOGIC := '1'; -- '1' when IP register will be created for a tag_threaded code interpreter
CONSTANT with_tokens       : STD_LOGIC := '1'; -- '1' when token threading will be supported. Token_width has to be set as well.

--------------------------------------------------------------------------
-- microcore busses
--------------------------------------------------------------------------

CONSTANT data_width        : NATURAL := 16; -- data bus width
CONSTANT inst_width        : NATURAL :=  8; -- instruction width

CONSTANT prog_addr_width   : NATURAL := 12; -- program memory address width
CONSTANT pcache_addr_width : NATURAL :=  0; -- prog. cache memory address width
CONSTANT prog_ram_width    : NATURAL :=  8; -- writeable program memory address width, at most data_width-2
                                            -- 0 => pure Harvard Architecture
CONSTANT ds_addr_width     : NATURAL :=  4; -- data stack pointer width
CONSTANT tasks_addr_width  : NATURAL :=  3; -- 2**tasks_addr_width copies of the stacks will be provided, one pair for each task

-- data memory parameters ****** data_addr_width > tasks_addr_width+rs_addr_width
CONSTANT data_addr_width   : NATURAL := 12; -- data memory address width, at most data_width-1 because of I/O space
CONSTANT dcache_addr_width : NATURAL :=  0; -- data cache memory address width, at most see "data_addr_width"
CONSTANT rs_addr_width     : NATURAL :=  8; -- return stack pointer width

CONSTANT token_width       : NATURAL :=  8; -- width of tokens of token threaded code - with_ip='1' needed
CONSTANT usr_vect_width    : NATURAL :=  3; -- each vector has room for 2**trap_vec_width instructions
CONSTANT reg_addr_width    : NATURAL :=  3; -- number of address bits reserved for internal registers at the top data space
CONSTANT interrupts        : NATURAL :=  2; -- maskable interrupt sources

SUBTYPE data_bus      IS STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);
SUBTYPE inst_bus      IS STD_LOGIC_VECTOR(inst_width-1 DOWNTO 0);
  SUBTYPE inst_type   IS STD_LOGIC_VECTOR(1 DOWNTO 0);
  SUBTYPE inst_stack  IS STD_LOGIC_VECTOR(1 DOWNTO 0);
  SUBTYPE inst_group  IS STD_LOGIC_VECTOR(inst_width-6 DOWNTO 0);

-- some synthesizers do not support SIGNALS in a package. Therefore, the INST-register
-- has to be routed into each entity explicitly as well as the field specific SIGNALs
  
SUBTYPE data_addr     IS STD_LOGIC_VECTOR(data_addr_width-1 DOWNTO 0);
SUBTYPE prog_addr     IS STD_LOGIC_VECTOR(prog_addr_width-1 DOWNTO 0);
SUBTYPE ds_addr       IS STD_LOGIC_VECTOR(ds_addr_width+tasks_addr_width-1 DOWNTO 0);
SUBTYPE rs_addr       IS STD_LOGIC_VECTOR(rs_addr_width+tasks_addr_width-1 DOWNTO 0);
SUBTYPE int_bus       IS STD_LOGIC_VECTOR(interrupts-1 DOWNTO 0);

--------------------------------------------------------------------------
-- microcore physical addresses and memory mapped registers
--------------------------------------------------------------------------

CONSTANT addr_reset   : prog_addr := to_vec(0*2**usr_vect_width, prog_addr_width);
CONSTANT addr_isr     : prog_addr := to_vec(1*2**usr_vect_width, prog_addr_width);
CONSTANT addr_esr     : prog_addr := to_vec(2*2**usr_vect_width, prog_addr_width);
CONSTANT addr_ovl     : prog_addr := to_vec(3*2**usr_vect_width, prog_addr_width);

CONSTANT FLAGS_REGISTER : data_addr := to_vec(-1, data_addr_width);
CONSTANT TASK_REGISTER  : data_addr := to_vec(-2, data_addr_width);
CONSTANT DEBUG_REGISTER : data_addr := to_vec(-3, data_addr_width);

--------------------------------------------------------------------------
-- status register
--------------------------------------------------------------------------

CONSTANT status_width : NATURAL :=  9;
CONSTANT s_c_bit      : NATURAL :=  0;  -- carry bit
CONSTANT s_ovl_bit    : NATURAL :=  1;  -- Overflow-bit of UDIVS instruction
CONSTANT s_ie_bit     : NATURAL :=  2;  -- Interrupt Enable bit
CONSTANT s_iis_bit    : NATURAL :=  3;  -- InterruptInService bit
CONSTANT s_lit_bit    : NATURAL :=  4;  -- LIT bit of the previous instruction
CONSTANT s_n_bit      : NATURAL :=  5;  -- Sign-bit of top data element (TOS or sometimes NOS)
CONSTANT s_z_bit      : NATURAL :=  6;  -- Zero-bit of top data element (TOS or sometimes NOS)
-- these are only active when debug='1'
CONSTANT s_in_bit     : NATURAL :=  7;  -- DEBUG_REG input ready
CONSTANT s_out_bit    : NATURAL :=  8;  -- DEBUG_REG output free

-- tos determines the flags in the status register when the status register is
-- written by STATUS! or IRET.
TYPE status_rec IS RECORD
           flags : STD_LOGIC_VECTOR(status_width-1 DOWNTO 0); -- status register flags, see above
           tos   : data_bus;
           write : STD_LOGIC; -- rewrite flags from TOS
END RECORD;

--------------------------------------------------------------------------
-- microcore record definitions
--------------------------------------------------------------------------
                     
TYPE  data_port  IS RECORD
                  en : STD_LOGIC;
               write : STD_LOGIC;
            internal : STD_LOGIC; -- internal register address space
          sel_dcache : STD_LOGIC; -- data cache address spaces
             sel_ram : STD_LOGIC; -- RAM address space, includes RSTACK
             sel_ext : STD_LOGIC; -- external I/O address space
                addr : data_addr;
                data : data_bus;  -- data on tri-state bus
END RECORD;

-- Control lines for Programm Memory
TYPE  program_port  IS RECORD
              write : STD_LOGIC;
               addr : prog_addr;
                din : inst_bus;
               dout : inst_bus;
END RECORD;

--------------------------------------------------------------------------
-- debug interface somewhat abstracted from Centronics interface
--------------------------------------------------------------------------

CONSTANT in_nibbles       : NATURAL :=  ceiling(data_width, 8);

-- Control lines for the umbilical
TYPE  debug_port  IS RECORD
            init     : STD_LOGIC;    -- program memory initialisation
            busy     : STD_LOGIC;    -- target is busy, don't expect attention

            in_str   : STD_LOGIC;    -- host placed valid Byte on DIN
            in_ack   : STD_LOGIC;    -- to host: last Byte read
            in_data  : STD_LOGIC_VECTOR(7 DOWNTO 0);

            out_str  : STD_LOGIC;    -- target: one word of data to send
            out_ack  : STD_LOGIC;    -- host: out_data has been captured
            out_data : STD_LOGIC;
END RECORD;

--------------------------------------------------------------------------
-- op codes
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- The opcodes' fields are defined in instruction register as follows:
--   7   6   5   4   3   2   1   0
-- |---|---|---|---|---|---|---|---|
-- |   |   |   |   |   |   |   |   |
-- |---|---|---|---|---|---|---|---|
-- |LIT|  TYPE | STACK |    GROUP  |
-- |---|-------|-------|-----------|
--
--------------------------------------------------------------------------
--
--------------------------------------------------------------------------
--                    TYPE
--------------------------------------------------------------------------
-- Code  Name  Action
--  00   BRA   Branches, Calls and Returns
--  01   ALU   Binary and Unary Operators
--  10   MEM   Data-Memory and Register access
--  11   USR   Unused by core, free for user extensions
--------------------------------------------------------------------------
CONSTANT op_BRA    : inst_type := "00";
CONSTANT op_ALU    : inst_type := "01";
CONSTANT op_MEM    : inst_type := "10";
CONSTANT op_USR    : inst_type := "11";

--------------------------------------------------------------------------
--                    STACK
--------------------------------------------------------------------------
-- Code  Name  Action
--  00   NONE  Type dependent
--  01   POP   Stack->NOS->TOS
--  10   PUSH  TOS->NOS->Stack
--  11   BOTH  Type dependent
--------------------------------------------------------------------------
CONSTANT op_NONE   : inst_stack := "00";
CONSTANT op_POP    : inst_stack := "01";
CONSTANT op_PUSH   : inst_stack := "10";
CONSTANT op_BOTH   : inst_stack := "11";

--------------------------------------------------------------------------------------
--                                  GROUPS
--------------------------------------------------------------------------------------
--  Code  Binary-Ops  Unary-Ops   Mathstep-Ops  Conditions    Branches       Registers
--  000      ADD        NOT         MULTS        NEVER          DUP            STATUS
--  001      ADC        SL          0DIVS        ALWAYS        ?DUP            TOR
--  010      SUB        ASR         UDIVS        ZERO           INT            RSTACK
--  011      SBC        LSR         DIVS         NZERO          EXC            LOCAL
--  100      AND        ROR         LDIVS        SIGN           OVL            RSP
--  101      OR         ROL                      NSIGN                         DSP
--  110      XOR        ZEQU        CRCS         CARRY          THREAD         TASK
--  111      NOS        CC          SWAP         NCARRY         TOKEN          IP
--------------------------------------------------------------------------------------

-- Binary-Operators
-- ALU POP and ALU PUSH
CONSTANT op_ADD    : inst_group := "000";
CONSTANT op_ADC    : inst_group := "001";
CONSTANT op_SUB    : inst_group := "010";
CONSTANT op_SSUB   : inst_group := "011";
CONSTANT op_AND    : inst_group := "100";
CONSTANT op_OR     : inst_group := "101";
CONSTANT op_XOR    : inst_group := "110";
CONSTANT op_NOS    : inst_group := "111";

-- Unary-Operators
-- ALU BOTH
CONSTANT op_NOT    : inst_group := "000";
CONSTANT op_SL     : inst_group := "001";
CONSTANT op_ASR    : inst_group := "010";
CONSTANT op_LSR    : inst_group := "011";
CONSTANT op_ROR    : inst_group := "100";
CONSTANT op_ROL    : inst_group := "101";
CONSTANT op_ZEQU   : inst_group := "110";
CONSTANT op_CC     : inst_group := "111";

-- Mathstep-Operators
-- ALU NONE
CONSTANT op_MULTS  : inst_group := "000";
CONSTANT op_0DIVS  : inst_group := "001";
CONSTANT op_UDIVS  : inst_group := "010";
CONSTANT op_DIVS   : inst_group := "011";
CONSTANT op_LDIVS  : inst_group := "100";
--CONSTANT op_       : inst_group := "101";
CONSTANT op_CRCS   : inst_group := "110";
CONSTANT op_SWAP   : inst_group := "111";

-- Conditions
-- BRA POP
CONSTANT op_NEVER  : inst_group := "000";
CONSTANT op_ALWAYS : inst_group := "001";
CONSTANT op_ZERO   : inst_group := "010";
CONSTANT op_NZERO  : inst_group := "011";
CONSTANT op_SIGN   : inst_group := "100";
CONSTANT op_NSIGN  : inst_group := "101";
CONSTANT op_NOVL   : inst_group := "110";
CONSTANT op_NCARRY : inst_group := "111";

-- Branches
-- BRA PUSH
CONSTANT op_DUP    : inst_group := "000";
CONSTANT op_EXC    : inst_group := "001";
CONSTANT op_QDUP   : inst_group := "010";
CONSTANT op_QOVL   : inst_group := "011";
CONSTANT op_INT    : inst_group := "100";
CONSTANT op_IRET   : inst_group := "101";
CONSTANT op_THREAD : inst_group := "110";
CONSTANT op_TOKEN  : inst_group := "111";

-- Registers
-- MEM NONE and MEM BOTH
CONSTANT op_STATUS : inst_group := "000";
CONSTANT op_TOR    : inst_group := "001";
CONSTANT op_RSTACK : inst_group := "010";
CONSTANT op_LOCAL  : inst_group := "011";
CONSTANT op_RSP    : inst_group := "100";
CONSTANT op_DSP    : inst_group := "101";
CONSTANT op_TASK   : inst_group := "110";
CONSTANT op_IP     : inst_group := "111";

CONSTANT NO_OP       : inst_bus := '0' & op_BRA & op_NONE & op_NEVER;
CONSTANT INT_OP      : inst_bus := '0' & op_BRA & op_PUSH & op_INT;
CONSTANT EXC_OP      : inst_bus := '0' & op_BRA & op_PUSH & op_EXC;
CONSTANT DROPFLAG_OP : inst_bus := '0' & op_BRA & op_POP  & op_NEVER;
CONSTANT NEXT_OP     : inst_bus := '0' & op_BRA & op_POP  & op_NZERO;
CONSTANT THREAD_OP   : inst_bus := '0' & op_BRA & op_PUSH & op_THREAD;
CONSTANT TOKEN_OP    : inst_bus := '0' & op_BRA & op_PUSH & op_TOKEN;
CONSTANT IPSTORE_OP  : inst_bus := '0' & op_MEM & op_NONE & op_IP;

END constants;
