-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore DSTACK.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- asynchronous Data Stack RAM to be compiled into the FPGA itself
-- using logic cells.
--
-- and
--
-- Data Stack entity consisting of the
-- TOS and NOS registers and the stack pointer DSP:

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY d_stack IS
PORT (reset     : IN    STD_LOGIC; 
      clk       : IN    STD_LOGIC;
      clk_en    : IN    STD_LOGIC;
      inst      : IN    inst_bus;
      uBus      : IN    data_port;
      alu       : IN    data_bus;
      tor       : IN    data_bus;
      ip_plus   : IN    data_bus;
      stBus     : INOUT status_rec;
      tos       : OUT   data_bus;
      lit       : OUT   data_bus;
      nos       : OUT   data_bus;
      stack     : OUT   data_bus;
      dsp       : OUT   data_bus);
END d_stack;

ARCHITECTURE rtl OF d_stack IS

SIGNAL i_lit      : STD_LOGIC;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;
SIGNAL nibble     : STD_LOGIC_VECTOR(inst_width-2 DOWNTO 0);
SIGNAL flags      : STD_LOGIC_VECTOR(status_width-1 DOWNTO 0);
SIGNAL s_lit      : STD_LOGIC;
SIGNAL s_n        : STD_LOGIC;
SIGNAL s_z        : STD_LOGIC;
SIGNAL sign       : STD_LOGIC;
SIGNAL zero       : STD_LOGIC;

SIGNAL tos_reg    : data_bus;
SIGNAL tos_new    : data_bus;
SIGNAL nos_reg    : data_bus;
SIGNAL nos_new    : data_bus;
SIGNAL stack_out  : data_bus;
SIGNAL uBus_data  : data_bus;
SIGNAL dsp_reg    : ds_addr;
SIGNAL dsp_new    : ds_addr;
SIGNAL dsp_plus   : ds_addr;
SIGNAL dsp_minus  : ds_addr;
SIGNAL stack_addr : ds_addr;
SIGNAL stack_wr   : STD_LOGIC;
SIGNAL status_wr  : STD_LOGIC;

BEGIN

lit     <= tos_new;
dsp     <= slice('0', data_width-(ds_addr_width+tasks_addr_width)) & dsp_reg;
nos     <= nos_reg;
tos     <= tos_reg;

i_lit <= inst(inst_width-1);
i_type <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);
nibble <= inst(inst_width-2 DOWNTO 0);
uBus_data <= uBus.data;
flags <= stBus.flags;
stack <= stack_out;

make_asyn_stack: IF  with_nos='1' AND syn_stackram='0'  GENERATE
data_stack: asynch_ram GENERIC MAP(data_width, (ds_addr_width+tasks_addr_width))
PORT MAP(clk   => clk,
         we    => stack_wr,
         addr  => stack_addr,
         di    => nos_reg,
         do    => stack_out);
END GENERATE make_asyn_stack;

make_nos_stack: IF  with_nos='1' AND syn_stackram='1' GENERATE
data_stack: internal_ram GENERIC MAP(data_width, (ds_addr_width+tasks_addr_width))
PORT MAP(clk   => clk,
         en    => clk_en,
         we    => stack_wr,
         addr  => dsp_new,
         di    => nos_reg,
         do    => stack_out);
END GENERATE make_nos_stack;

make_dp_stack: IF  with_nos='0'  GENERATE
data_stack: dualport_ram GENERIC MAP(data_width, (ds_addr_width+tasks_addr_width))
PORT MAP(clk   => clk,
         we    => stack_wr,
         waddr => dsp_new,
         raddr => stack_addr,
         di    => nos_new,
         do    => nos_reg);
stack_out <= nos_reg;
END GENERATE make_dp_stack;

dsp_moves : PROCESS (dsp_reg)
BEGIN
  IF  tasks_addr_width=0  THEN
dsp_plus <= dsp_reg+1;
dsp_minus <= dsp_reg-1;
  ELSE
     dsp_plus <= dsp_reg(tasks_addr_width+ds_addr_width-1 DOWNTO ds_addr_width) & dsp_reg(ds_addr_width-1 DOWNTO 0)+1;
     dsp_minus <= dsp_reg(tasks_addr_width+ds_addr_width-1 DOWNTO ds_addr_width) & dsp_reg(ds_addr_width-1 DOWNTO 0)-1;
  END IF;
END PROCESS dsp_moves;

------------------------------------------------------------------------------
-- dstack_control: Generates control lines for data stack entity
-- The opcodes' fields are defined in instruction register as follows:
--   7   6   5   4   3   2   1   0
-- |---|---|---|---|---|---|---|---|
-- |   |   |   |   |   |   |   |   |
-- |---|---|---|---|---|---|---|---|
-- |LIT|  TYPE | STACK |    GROUP  |
-- |---|-------|-------|-----------|
--
-- In the process the op-field are preceded by "i_"
-- The different "op_XXX" are defined in constants.vhdl
-- s_lit is the actual value of LIT in STATUS
-- 
------------------------------------------------------------------------------
dstack_control: PROCESS (nos_reg, tos_reg, i_lit, i_type, i_stack, i_group, flags, zero, stack_out, dsp_reg, uBus_data, alu, s_lit, nibble, dsp_plus, dsp_minus, tor, ip_plus)

PROCEDURE push_stack (i : IN data_bus) IS
BEGIN
  nos_new    <= i;
  dsp_new    <= dsp_plus;
  stack_addr <= dsp_plus;
  stack_wr   <= '1';
END push_stack;

PROCEDURE pop_stack IS
BEGIN
  nos_new    <= stack_out;
  dsp_new    <= dsp_minus;
  stack_addr <= dsp_reg;
  stack_wr   <= '0';
END pop_stack;

BEGIN

tos_new    <= tos_reg;
nos_new    <= nos_reg;
dsp_new    <= dsp_reg;
stack_addr <= dsp_reg;
stack_wr   <= '0';
status_wr  <= '0';

IF  i_lit='0'  THEN
   CASE  i_type  IS
   WHEN op_BRA =>
      CASE i_stack IS
      WHEN op_NONE =>
         CASE i_group IS
         WHEN op_ZERO  => -- 0=EXIT, IRET
            pop_stack;
            tos_new <= nos_reg;
         WHEN op_NZERO =>  -- ?EXIT
            IF  with_tor='1'  THEN
               pop_stack;
               tos_new <= nos_reg;
            ELSE
               pop_stack;
               tos_new <= nos_reg;
            END IF;
         WHEN OTHERS => NULL;
         END CASE;
      WHEN op_POP  => pop_stack;
                      tos_new <= nos_reg;
                      IF  i_group=op_NEVER  THEN  -- drop_flag instruction
                         tos_new <= tos_reg;
                      END IF;
      WHEN op_PUSH =>
         CASE i_group IS
         WHEN op_DUP    => push_stack(tos_reg);
         WHEN op_EXC    => NULL;
         WHEN op_QOVL   => NULL;
         WHEN op_QDUP   => IF  zero='0'  THEN -- ?DUP
                              push_stack(tos_reg);
                           END IF;
         WHEN op_INT    => push_stack(tos_reg);
                           tos_new <= slice('0',data_width-status_width) & flags;
         WHEN op_IRET   => pop_stack;
                           tos_new <= nos_reg;
                           status_wr <= '1';
         WHEN op_THREAD => IF  with_ip='1'  THEN   -- THREAD
                              IF  uBus_data(data_width-1)='0'  THEN
                                 push_stack(tos_reg);
                                 tos_new <= ip_plus;
                              END IF;
                           END IF;
         WHEN op_TOKEN  => IF  with_ip='1' AND with_tokens='1'  THEN  -- TOKEN
                              pop_stack;
                              tos_new <= nos_reg;
                           END IF;
         WHEN OTHERS => push_stack(tos_reg);
         END CASE;
      WHEN OTHERS  => -- op_BOTH
         pop_stack;
         tos_new <= nos_reg;
      END CASE;
   WHEN op_ALU =>
      tos_new <= alu;
      CASE i_stack IS
      WHEN op_NONE =>
         CASE i_group IS
         WHEN op_MULTS => IF  with_tor='1' AND tor(tor'high)='0'  THEN
                             tos_new <= tos_reg(tos_reg'high-1 DOWNTO 0) & '0';
                          END IF;
         WHEN op_0DIVS => IF  with_tor='1' AND with_nos='1'  THEN
                             nos_new <= tos_reg;
                             dsp_new <= dsp_minus; -- remove dividend_low from dstack
                          END IF;
         WHEN op_SWAP  => IF  with_nos='1'  THEN  nos_new <= tos_reg;  END IF;
         WHEN OTHERS   => NULL;
         END CASE;
      WHEN op_POP  => pop_stack;
                      IF  i_group=op_NOS  THEN
                         tos_new <= tos_reg; -- NIP instruction
                      END IF;
      WHEN op_PUSH => push_stack(tos_reg);
      WHEN OTHERS  => NULL;
      END CASE;
   WHEN op_MEM =>
      CASE i_stack IS
      WHEN op_NONE =>
         IF  i_group=op_DSP  THEN -- dsp!
               dsp_new <= uBus_data(tasks_addr_width+ds_addr_width-1 DOWNTO 0);
            stack_addr <= uBus_data(tasks_addr_width+ds_addr_width-1 DOWNTO 0);
         ELSE
            pop_stack;
            tos_new <= nos_reg;
            CASE  i_group  IS
            WHEN op_STATUS => status_wr <= '1';
            WHEN op_LOCAL  => IF  with_locals='1' THEN  tos_new <= alu;  END IF;
            WHEN op_TASK   => IF  with_tasks='1'  THEN  tos_new <= alu;  END IF;
            WHEN OTHERS => NULL;
            END CASE;
         END IF;
      WHEN op_POP  =>
         pop_stack;
         tos_new <= alu;
      WHEN op_PUSH =>
         push_stack(uBus_data);
         tos_new <= alu;
      WHEN OTHERS  =>  -- op_BOTH
         push_stack(tos_reg);
         tos_new <= uBus_data;
         CASE i_group IS
         WHEN op_TOR | op_RSTACK => IF  with_tor='1'  THEN  tos_new <= tor;  END IF;
         WHEN op_LOCAL => IF  with_locals='1'  THEN
                             push_stack(uBus_data);
                             tos_new <= alu;
                          END IF;
         WHEN op_TASK =>  IF  with_tasks='1'  THEN
                             push_stack(uBus_data);
                             tos_new <= alu;
                          END IF;
         WHEN OTHERS => NULL;
         END CASE;
      END CASE;
   WHEN OTHERS => NULL; -- op_USR
   END CASE;
ELSE -- i_lit='1'
   tos_new <= tos_reg(data_width-8 DOWNTO 0) & nibble;
   IF  s_lit='0'  THEN
      push_stack(tos_reg);
      tos_new <= slice(nibble(6), data_width-(inst_width-1)) & nibble;
   END IF;
END IF;

END PROCESS dstack_control;

stack_registers: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      tos_reg <= (OTHERS => '0');
      dsp_reg <= (OTHERS => '0');
      IF  with_nos='1'  THEN
         nos_reg <= (OTHERS => '0');
      END IF;
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
        tos_reg <= tos_new;
        dsp_reg <= dsp_new;
        IF  with_nos='1'  THEN
           nos_reg <= nos_new;
        END IF;
     END IF;
   END IF;
END PROCESS stack_registers;

------------------------------------------------------------------------------
-- sign_and_zero: Fetch the tos status
--             STATUS 
--   7   6   5   4   3   2   1   0
-- |---|---|---|---|---|---|---|---|
-- |   |   |   |   |   |   |   |   |
-- |---|---|---|---|---|---|---|---|
-- |   |   | Z | N |LIT|IIS|IE | C |
-- |---|---|---|---|---|---|---|---|
--
-- (a) reset = 1: N, Z and LIT all 0 
-- (b) clk valid & status.write = 1: fetch flags from tos
-- (c) clk valid & LIT = 0: signals sign and zero determine N and Z flags
-- (d) clk valid & s_lit = 0: signals sign and zero determine N and Z flags
------------------------------------------------------------------------------

sign <= tos_reg(data_width-1);
zero <= '1' WHEN  tos_reg=slice('0', data_width)  ELSE  '0';

sign_and_zero: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      s_n <= '0';
      s_z <= '0';
      s_lit <= '0';
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
       s_lit <= i_lit;
       IF  stBus.write='1'  THEN
          s_n   <= stBus.tos(s_n_bit);
          s_z   <= stBus.tos(s_z_bit);
          s_lit <= stBus.tos(s_lit_bit);
          IF  stBus.tos(s_lit_bit)='0'  THEN
             s_n <= sign;
             s_z <= zero;
          END IF;
       ELSIF  s_lit='0'  THEN
          s_n <= sign;
          s_z <= zero;
       END IF;
       IF  inst=DROPFLAG_OP  THEN
          s_lit <= s_lit;
       END IF;
     END IF;
   END IF;
END PROCESS sign_and_zero;

stBus.flags(s_lit_bit) <= s_lit;
stBus.flags(s_n_bit)   <= sign  WHEN  s_lit='0'  ELSE  s_n;
stBus.flags(s_z_bit)   <= zero  WHEN  s_lit='0'  ELSE  s_z;
stBus.write            <= status_wr;
stBus.tos              <= tos_reg;

END rtl;

