-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore FUNCTIONS.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- Functions and components which are used throughout MicroCore.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_arith.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;

PACKAGE functions IS

FUNCTION  to_int(v : IN STD_LOGIC_VECTOR ) RETURN INTEGER;

FUNCTION  to_vec(v : IN INTEGER;
                 s : IN INTEGER          ) RETURN STD_LOGIC_VECTOR;

FUNCTION   slice(v : IN STD_LOGIC;
                 s : IN INTEGER          ) RETURN STD_LOGIC_VECTOR;

FUNCTION     crc(d : IN STD_LOGIC_VECTOR;
                 p : IN STD_LOGIC_VECTOR ) RETURN STD_LOGIC_VECTOR;

FUNCTION     max(v : IN INTEGER;
                 w : IN INTEGER          ) RETURN INTEGER;

FUNCTION    log2(v : IN INTEGER          ) RETURN INTEGER;

FUNCTION ceiling(v : IN INTEGER;
                 w : IN INTEGER          ) RETURN INTEGER;

FUNCTION   adder(v : IN STD_LOGIC_VECTOR;
                 w : IN STD_LOGIC_VECTOR;
                 carry : IN STD_LOGIC    ) RETURN STD_LOGIC_VECTOR;

ATTRIBUTE syn_noprune : BOOLEAN;

COMPONENT pullup
PORT(o : OUT STD_LOGIC := 'H');
END COMPONENT;
ATTRIBUTE syn_noprune OF pullup : COMPONENT IS true;

COMPONENT pulldown
PORT(o : OUT STD_LOGIC := 'L');
END COMPONENT;
ATTRIBUTE syn_noprune OF pulldown : COMPONENT IS true;

COMPONENT synchronise
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;

      async    : IN    STD_LOGIC;
      sync     : OUT   STD_LOGIC);
END COMPONENT;

COMPONENT synchronise_n
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;
      async_n  : IN    STD_LOGIC;
      sync     : OUT   STD_LOGIC);
END COMPONENT;

COMPONENT debounce
GENERIC (width : INTEGER);
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;
      async    : IN    STD_LOGIC;
      sync     : OUT   STD_LOGIC);
END COMPONENT;

COMPONENT debounce_n
GENERIC (width : INTEGER);
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;
      async    : IN    STD_LOGIC;
      sync     : OUT   STD_LOGIC);
END COMPONENT;

COMPONENT edge
     PORT (reset   : IN  STD_LOGIC;
           clk     : IN  STD_LOGIC;
           i       : IN  STD_LOGIC;
           o       : OUT STD_LOGIC);
END COMPONENT;

COMPONENT edge_n
     PORT (reset   : IN  STD_LOGIC;
           clk     : IN  STD_LOGIC;
           i       : IN  STD_LOGIC;
           o       : OUT STD_LOGIC);
END COMPONENT;

COMPONENT external_ram
GENERIC (data_width : INTEGER;
         addr_width : INTEGER);
PORT (enable : IN    STD_LOGIC;
      write  : IN    STD_LOGIC;
      addr   : IN    STD_LOGIC_VECTOR;
      data   : INOUT STD_LOGIC_VECTOR);
END COMPONENT;

COMPONENT asynch_ram
GENERIC (data_width : INTEGER;
         addr_width : INTEGER);
PORT (clk   : IN    STD_LOGIC;
      we    : IN    STD_LOGIC;
      addr  : IN    STD_LOGIC_VECTOR;
      di    : IN    STD_LOGIC_VECTOR;
      do    : OUT   STD_LOGIC_VECTOR);
END COMPONENT;

COMPONENT internal_ram
GENERIC (data_width : INTEGER;
         addr_width : INTEGER);
PORT (clk   : IN    STD_LOGIC;
      en    : IN    STD_LOGIC;
      we    : IN    STD_LOGIC;
      addr  : IN    STD_LOGIC_VECTOR;
      di    : IN    STD_LOGIC_VECTOR;
      do    : OUT   STD_LOGIC_VECTOR);
END COMPONENT;

COMPONENT dualport_ram
GENERIC (data_width  : INTEGER;
         addr_width  : INTEGER);
PORT (clk   : IN    STD_LOGIC;
      we    : IN    STD_LOGIC;
      waddr : IN    STD_LOGIC_VECTOR;
      raddr : IN    STD_LOGIC_VECTOR;
      di    : IN    STD_LOGIC_VECTOR;
      do    : OUT   STD_LOGIC_VECTOR);
END COMPONENT;

END functions;

PACKAGE BODY functions IS

    ----------------------------------------------------------------
    -- to_int
    -- Convert STD_LOGIC_VECTOR to INTEGER number 
    ----------------------------------------------------------------

    FUNCTION to_int(v  :  IN STD_LOGIC_VECTOR
                   )  RETURN INTEGER IS
        VARIABLE temp  :     INTEGER;
    BEGIN
        temp := conv_INTEGER(v);
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- to_vec
    -- Convert INTEGER number to STD_LOGIC_VECTOR
    ----------------------------------------------------------------

    FUNCTION to_vec(v: IN INTEGER;
                    s  :  IN INTEGER
                   )  RETURN STD_LOGIC_VECTOR IS
        VARIABLE temp  : STD_LOGIC_VECTOR(s-1 DOWNTO 0);
    BEGIN
       temp := conv_STD_LOGIC_VECTOR(v,s);
       RETURN temp;
    END;

    ----------------------------------------------------------------
    -- slice 
    -- Return (others => v) within range s-1 to 0
    ----------------------------------------------------------------

    FUNCTION slice(v : IN STD_LOGIC;
                   s   :  IN INTEGER
                  )   RETURN STD_LOGIC_VECTOR IS
       VARIABLE temp : STD_LOGIC_VECTOR (s-1 DOWNTO 0);
    BEGIN
        temp := (OTHERS => v); -- synopsys compatibility
       RETURN temp;
    END;

    ----------------------------------------------------------------
    -- crc-function (not for synthesis)
    ----------------------------------------------------------------

    FUNCTION crc(d : IN STD_LOGIC_VECTOR;
                 p     :  IN STD_LOGIC_VECTOR
                )     RETURN STD_LOGIC_VECTOR IS
       VARIABLE temp : STD_LOGIC_VECTOR (p'HIGH DOWNTO p'LOW);
    BEGIN
       temp := (OTHERS => '0');
       FOR I IN d'HIGH DOWNTO  d'LOW LOOP
           IF   d(I) = temp(p'HIGH) THEN
               temp := temp(p'HIGH - 1 DOWNTO p'LOW) & '0';
           ELSE
               temp := temp(p'HIGH - 1 DOWNTO p'LOW) & '0';
               temp := temp XOR p;
           END IF;
       END LOOP;
       RETURN temp;
    END;

    ----------------------------------------------------------------
    -- max
    -- Return maximum of two INTEGERs w and v
    ----------------------------------------------------------------

    FUNCTION max(v  :  IN INTEGER;
                 w  :  IN INTEGER
                )  RETURN INTEGER IS
    BEGIN
        IF (v > w) THEN  RETURN v;  ELSE  RETURN w;  END IF;
    END;

    ----------------------------------------------------------------
    -- logarithm dualis
    ----------------------------------------------------------------

    FUNCTION log2(v    :  IN INTEGER
                 )    RETURN INTEGER IS
        VARIABLE temp  :     INTEGER;
    BEGIN
        temp := 0;
        WHILE  2 ** temp < v  LOOP
            temp := temp + 1;
        END LOOP;
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- ceiling of two number
    ----------------------------------------------------------------

    FUNCTION ceiling(v : IN INTEGER;
                     w : IN INTEGER
                    ) RETURN INTEGER IS
        VARIABLE temp  : INTEGER;
    BEGIN
        temp := 1;
        WHILE  temp * w < v  LOOP
            temp := temp + 1;
        END LOOP;
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- adder of the alu
    ----------------------------------------------------------------

    FUNCTION adder(v     : IN  STD_LOGIC_VECTOR;
                   w     : IN  STD_LOGIC_VECTOR;
                   carry : IN  STD_LOGIC
                  ) RETURN STD_LOGIC_VECTOR IS
        VARIABLE temp : STD_LOGIC_VECTOR (v'HIGH + 1 DOWNTO v'LOW);
    BEGIN
        temp := ('0' & v) + ('0' & w) + carry;
        RETURN temp;
    END;

END functions;

----------------------------------------------------------------
----------------------------------------------------------------
----------------------------------------------------------------
-- synchronising asynchronous input signals
----------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;

ENTITY synchronise IS
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;

      async    : IN    STD_LOGIC; 
      sync     : OUT   STD_LOGIC);
END synchronise;

ARCHITECTURE rtl OF synchronise IS
BEGIN

synch: PROCESS(clk, reset)
BEGIN
  IF  reset='1'  THEN
    sync <= '0';
  ELSIF  rising_edge(clk)  THEN
    sync <= async;
  END IF;
END PROCESS synch; 

END rtl;

----------------------------------------------------------------
-- synchronising inverted asynchronous input signals
----------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;

ENTITY synchronise_n IS
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;
      async_n  : IN    STD_LOGIC; 
      sync     : OUT   STD_LOGIC);
END synchronise_n;

ARCHITECTURE rtl OF synchronise_n IS
BEGIN

synch_n: PROCESS(clk, reset)
BEGIN
  IF  reset='1'  THEN
    sync <= '0';
  ELSIF  rising_edge(clk)  THEN
    sync <= NOT async_n;
  END IF;
END PROCESS synch_n; 

END rtl;

----------------------------------------------------------------
-- debouncing asynchronous input signals
----------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;

ENTITY debounce IS
GENERIC (width : INTEGER := 3);
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;
      async    : IN    STD_LOGIC;
      sync     : OUT   STD_LOGIC);
END debounce;

ARCHITECTURE rtl OF debounce IS
SIGNAL temp : STD_LOGIC_VECTOR(width-2 DOWNTO 0);
BEGIN

noglitch: PROCESS(clk, reset)
BEGIN
  IF  reset='1'  THEN
    sync <= '0';
    temp <= (OTHERS => '0');
  ELSIF  rising_edge(clk)  THEN
    temp <= temp(temp'high-1 DOWNTO 0) & async;
    IF  (temp&async)=-1  THEN
       sync <= '1';
    END IF;
    IF  (temp&async)=0  THEN
       sync <= '0';
    END IF;
  END IF;
END PROCESS noglitch; 

END rtl;

----------------------------------------------------------------
-- debouncing inverted asynchronous input signals
----------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;

ENTITY debounce_n IS
GENERIC (width : INTEGER := 3);
PORT (reset    : IN    STD_LOGIC;
      clk      : IN    STD_LOGIC;
      async    : IN    STD_LOGIC;
      sync     : OUT   STD_LOGIC);
END debounce_n;

ARCHITECTURE rtl OF debounce_n IS
SIGNAL temp : STD_LOGIC_VECTOR(width-2 DOWNTO 0);
BEGIN

noglitch_n: PROCESS(clk, reset)
BEGIN
  IF  reset='1'  THEN
    sync <= '0';
    temp <= (OTHERS => '0');
  ELSIF  rising_edge(clk)  THEN
    temp <= temp(temp'high-1 DOWNTO 0) & async;
    IF  (temp&async)=-1  THEN
       sync <= '0';
    END IF;
    IF  (temp&async)=0  THEN
       sync <= '1';
    END IF;
  END IF;
END PROCESS noglitch_n;

END rtl;

----------------------------------------------------------------
-- metastable safe positive edge detector
----------------------------------------------------------------

Library IEEE;
USE IEEE.STD_LOGIC_1164.ALL;

ENTITY edge IS
     PORT (reset   : IN  STD_LOGIC;
           clk     : IN  STD_LOGIC;
           i       : IN  STD_LOGIC;
           o       : OUT STD_LOGIC);
END edge;

ARCHITECTURE rtl OF edge IS

SIGNAL temp   : STD_LOGIC;
SIGNAL output : STD_LOGIC;

BEGIN

o <= output;

risingedge: PROCESS(reset, clk)
BEGIN
   IF  reset='1'  THEN
      temp   <= '0';          -- temp
      output <= '0';          -- out
   ELSIF  rising_edge(clk)  THEN
      temp   <= i AND     (temp OR output);
      output <= i AND NOT (temp OR output);
   END IF;
END PROCESS;

END rtl;

----------------------------------------------------------------
-- metastable safe negative edge detector
----------------------------------------------------------------

Library IEEE;
USE IEEE.STD_LOGIC_1164.ALL;

ENTITY edge_n IS
     PORT (reset   : IN  STD_LOGIC;
           clk     : IN  STD_LOGIC;
           i       : IN  STD_LOGIC;
           o       : OUT STD_LOGIC);
END edge_n;

ARCHITECTURE rtl OF edge_n IS

SIGNAL temp   : STD_LOGIC;
SIGNAL output : STD_LOGIC;

BEGIN

o   <= output;

fallingedge: PROCESS(reset, clk)
BEGIN
   IF reset = '1' THEN
      temp   <= '1';          -- temp
      output <= '0';          -- out
   ELSIF  rising_edge(clk)  THEN
      temp   <= NOT i AND     (temp OR output);
      output <= NOT i AND NOT (temp OR output);
   END IF;
END PROCESS;

END rtl;

-----------------------------------------------------------------------
-- external_ram: Input data is stored on falling WRITE edge
-----------------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_unsigned.ALL;

ENTITY external_ram IS
GENERIC (data_width  : INTEGER;
         addr_width  : INTEGER);
PORT (enable : IN    STD_LOGIC;
      write  : IN    STD_LOGIC;
      addr   : IN    STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);
      data   : INOUT STD_LOGIC_VECTOR(data_width-1 DOWNTO 0));
END external_ram;

ARCHITECTURE sim_model OF external_ram IS

TYPE ram_type IS ARRAY (2**addr_width-1 DOWNTO 0) OF STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);

SIGNAL addr_i  : STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);
SIGNAL din     : STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);

BEGIN

addr_i <= addr AFTER 1 ns; -- delay for the sake of simulation
din <= data AFTER 1 ns; -- delay for the sake of simulation

external_ram: PROCESS(enable, write, addr_i)
VARIABLE ram : ram_type;
BEGIN
  data <= (OTHERS => 'Z');
  IF  enable='1'  THEN
     IF  falling_edge(write)  THEN
        ram(conv_integer(addr_i)) := din;
     END IF;
     IF  write='0'  THEN
        data <= ram(conv_integer(addr_i));
     END IF;
  END IF;
END PROCESS external_ram;

END sim_model;

-----------------------------------------------------------------------
-- internal asynchronous RAM made of LUTs
-----------------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_unsigned.ALL;

ENTITY asynch_ram IS
GENERIC (data_width  : INTEGER;
         addr_width  : INTEGER);
PORT (clk   : IN  STD_LOGIC;
      we    : IN  STD_LOGIC;
      addr  : IN  STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);
      di    : IN  STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);
      do    : OUT STD_LOGIC_VECTOR(data_width-1 DOWNTO 0));
END asynch_ram;

ARCHITECTURE inference_model OF asynch_ram IS

TYPE ram_type IS ARRAY (2**addr_width-1 DOWNTO 0) OF STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);

SIGNAL ram : ram_type;

BEGIN

asyn_ram: PROCESS(clk)
BEGIN
  IF  rising_edge(clk)  THEN
     IF  we='1'  THEN
        ram(conv_integer(addr)) <= di ;
     END IF;
  END IF;
END PROCESS asyn_ram;

do <= ram(conv_integer(addr));

END inference_model;

-----------------------------------------------------------------------
-- internal_blockRAM: Address latched prior to access,
--               Input data stored on rising clock edge
-----------------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_unsigned.ALL;

ENTITY internal_ram IS
GENERIC (data_width  : INTEGER;
         addr_width  : INTEGER);
PORT (clk   : IN    STD_LOGIC;
      en    : IN    STD_LOGIC;
      we    : IN    STD_LOGIC;
      addr  : IN    STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);
      di    : IN    STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);
      do    : OUT   STD_LOGIC_VECTOR(data_width-1 DOWNTO 0));

ATTRIBUTE ram_style : STRING;
ATTRIBUTE ram_style OF internal_ram : ENTITY IS "block";

END internal_ram;

ARCHITECTURE inference_model OF internal_ram IS

ATTRIBUTE syn_ramstyle : STRING;

TYPE ram_type IS ARRAY (2**addr_width-1 DOWNTO 0) OF STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);

SIGNAL ram : ram_type; ATTRIBUTE syn_ramstyle OF ram : SIGNAL IS "block_ram";

BEGIN

internal_ram: PROCESS(clk)
BEGIN
  IF  rising_edge(clk)  THEN
     IF  en='1'  THEN
        IF  we='1'  THEN
           ram(conv_integer(addr)) <= di;
           do <= di;
        ELSE
           do <= ram(conv_integer(addr));
        END IF;
     END IF;
  END IF;
END PROCESS internal_ram;

END inference_model;

-----------------------------------------------------------------------
-- internal_dual-port blockRAM: 
--    Address latched prior to access,
--    Input data stored on rising clock edge
-----------------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_unsigned.ALL;

ENTITY dualport_ram IS
GENERIC (data_width  : INTEGER;
         addr_width  : INTEGER);
PORT (clk   : IN    STD_LOGIC;
      we    : IN    STD_LOGIC;
      waddr : IN    STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);
      raddr : IN    STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);
      di    : IN    STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);
      do    : OUT   STD_LOGIC_VECTOR(data_width-1 DOWNTO 0));

ATTRIBUTE ram_style : STRING;
ATTRIBUTE ram_style OF dualport_ram : ENTITY IS "block";

END dualport_ram;

ARCHITECTURE inference_model OF dualport_ram IS

ATTRIBUTE syn_ramstyle : STRING;

TYPE ram_type IS ARRAY (2**addr_width-1 DOWNTO 0) OF STD_LOGIC_VECTOR(data_width-1 DOWNTO 0);

SIGNAL ram     : ram_type; ATTRIBUTE syn_ramstyle OF ram : SIGNAL IS "block_ram";
SIGNAL raddr_d : STD_LOGIC_VECTOR(addr_width-1 DOWNTO 0);

BEGIN

dual_port: PROCESS(clk)
BEGIN
  IF  rising_edge(clk)  THEN
     IF  we='1'  THEN
        ram(conv_integer(waddr)) <= di;
     END IF;
     raddr_d <= raddr;
  END IF;
END PROCESS dual_port;

do <= ram(conv_integer(raddr_d));

END inference_model;
