-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore UBUS3STATE.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- internal bus realised using 3-state buffers.
-- To prevent bus contention, signal UBUS.EN is used to switch off all bus
-- drivers for the first quarter of each instruction execution cycle.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY microbus IS
PORT (reset     : IN    STD_LOGIC;
      clk       : IN    STD_LOGIC;
      clk_en    : IN    STD_LOGIC;
      inst      : IN    inst_bus;
      stBus     : IN    status_rec;
      condition : IN    STD_LOGIC;
      token     : IN    STD_LOGIC;
      pc        : IN    data_bus;
      tos       : IN    data_bus;
      nos       : IN    data_bus;
      alu       : IN    data_bus;
      dsp       : IN    data_bus;
      rsp       : IN    data_bus;
      tor       : IN    data_bus;
      flags     : IN    data_bus;
      ip        : IN    data_bus;
      sel_data  : IN    STD_LOGIC;
      uBus      : INOUT data_port;
      romdata   : IN    inst_bus;
      sel_rom1  : OUT   STD_LOGIC;
      sel_rom2  : OUT   STD_LOGIC;
      task      : OUT   data_bus);
END microbus;

ARCHITECTURE tri_state OF microbus IS

ATTRIBUTE syn_keep : BOOLEAN;

SIGNAL i_lit      : STD_LOGIC;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;
SIGNAL s_lit      : STD_LOGIC;
SIGNAL status_sel : STD_LOGIC; ATTRIBUTE syn_keep OF status_sel : SIGNAL IS true;
SIGNAL rsp_sel    : STD_LOGIC; ATTRIBUTE syn_keep OF rsp_sel    : SIGNAL IS true;
SIGNAL tor_sel    : STD_LOGIC; ATTRIBUTE syn_keep OF tor_sel    : SIGNAL IS true;
SIGNAL dsp_sel    : STD_LOGIC; ATTRIBUTE syn_keep OF dsp_sel    : SIGNAL IS true;
SIGNAL flags_sel  : STD_LOGIC; ATTRIBUTE syn_keep OF flags_sel  : SIGNAL IS true;
SIGNAL tos_sel    : STD_LOGIC; ATTRIBUTE syn_keep OF tos_sel    : SIGNAL IS true;
SIGNAL nos_sel    : STD_LOGIC; ATTRIBUTE syn_keep OF nos_sel    : SIGNAL IS true;
SIGNAL pc_sel     : STD_LOGIC; ATTRIBUTE syn_keep OF pc_sel     : SIGNAL IS true;
SIGNAL ip_sel     : STD_LOGIC; ATTRIBUTE syn_keep OF ip_sel     : SIGNAL IS true;
SIGNAL none_sel   : STD_LOGIC; ATTRIBUTE syn_keep OF none_sel   : SIGNAL IS true;

SIGNAL write      : STD_LOGIC;
SIGNAL internal   : STD_LOGIC;
SIGNAL sel_ram    : STD_LOGIC;
SIGNAL sel_ext    : STD_LOGIC;
SIGNAL iospace    : STD_LOGIC;
SIGNAL romspace   : STD_LOGIC;
SIGNAL regspace   : STD_LOGIC;
SIGNAL dcache     : STD_LOGIC;
SIGNAL rom_sel    : STD_LOGIC;
SIGNAL rom_sel_d  : STD_LOGIC;

SIGNAL task_i     : data_addr;
SIGNAL task_sel   : STD_LOGIC; ATTRIBUTE syn_keep OF task_sel : SIGNAL IS true;

BEGIN

uBus.write    <= write;
uBus.internal <= internal;
uBus.sel_dcache <= sel_ram AND dcache;
uBus.sel_ram  <= sel_ram AND NOT dcache;
uBus.sel_ext  <= sel_ext;
sel_rom1 <= rom_sel;
sel_rom2 <= rom_sel_d;
task <= slice('0', data_width-data_addr_width) & task_i;

i_lit    <= inst(inst_width-1);
i_type   <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack  <= inst(inst_width-4 DOWNTO inst_width-5);
i_group  <= inst(inst_width-6 DOWNTO 0);
s_lit    <= stBus.flags(s_lit_bit);

------------------------------------------------------------------------------
-- tri_state bus 
-- The simulation states under which conditions the contents of tos, nos, 
-- dsp, rsp, alu, data and status are transferred via the microbus.
-- The conditions are so contrieved that only one statement will be
-- true in the time unit considered.
--
-- >>>>>> IMPORTANT NOTICE <<<<<<<<< 
--
-- Signal uBus.en has been added which switches off all bus drivers at the
-- beginning of a new cycle such that no bus contention will occur
-- due to delay differences in generating the control signals.
-- uBus.en is derived from SYSCLK, which must be at least 2*uClk
------------------------------------------------------------------------------

-- pragma resolution_method three_state
uBus.data <= dsp        WHEN  uBus.en='1' AND dsp_sel='1'                   ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= rsp        WHEN  uBus.en='1' AND rsp_sel='1'                   ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= tor        WHEN  uBus.en='1' AND tor_sel='1' AND with_tor='1'  ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= flags      WHEN  uBus.en='1' AND flags_sel='1'                 ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= slice('0', data_width-status_width) & stBus.flags
                        WHEN  uBus.en='1' AND status_sel='1'                ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= tos        WHEN  uBus.en='1' AND tos_sel='1'                   ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= nos        WHEN  uBus.en='1' AND nos_sel='1'                   ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= pc         WHEN  uBus.en='1' AND pc_sel='1'                    ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= slice('0', data_width-inst_width) & romdata
                        WHEN  uBus.en='1' AND rom_sel='1'  AND write='0'    ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= slice('0', data_width-data_addr_width) & task_i    
                        WHEN  uBus.en='1' AND task_sel='1' AND write='0'    ELSE  (OTHERS => 'Z') AFTER 5 ns;
uBus.data <= ip         WHEN  uBus.en='1' AND ip_sel='1' AND write='0'      ELSE  (OTHERS => 'Z') AFTER 5 ns;
--uBus.data <= uBus.data  WHEN  uBus.en='1' AND none_sel='1'                  ELSE  (OTHERS => 'Z') AFTER 5 ns;

--none_sel <= NOT(dsp_sel OR rsp_sel OR ip_sel OR flags_sel OR status_sel OR tos_sel OR nos_sel OR pc_sel OR rom_sel OR task_sel OR sel_data);

iospace  <= alu(data_width-1);
regspace <= '1' WHEN  alu(data_addr_width   DOWNTO reg_addr_width)=-1  ELSE '0';
romspace <= '1' WHEN  alu(data_addr_width-1 DOWNTO prog_ram_width)=0    AND prog_ram_width/=0     ELSE '0';
dcache   <= '1' WHEN  data_addr_width=dcache_addr_width                                           ELSE
            '1' WHEN  uBus.addr(data_addr_width-1 DOWNTO dcache_addr_width)=0 AND dcache_addr_width/=0  ELSE
            '0';

flags_sel <= '1' WHEN  uBus.addr=FLAGS_REGISTER AND internal='1' AND write='0'  ELSE '0';

uBus_control: PROCESS (i_lit, i_type, i_stack, i_group, s_lit, condition, regspace, iospace, romspace, rom_sel_d, token)
BEGIN

status_sel <= '0';
rsp_sel    <= '0';
tor_sel    <= '0';
dsp_sel    <= '0';
ip_sel     <= '0';
tos_sel    <= '0';
nos_sel    <= '0';
pc_sel     <= '0';
write      <= '0';
internal   <= '0';
sel_ram    <= '0';
sel_ext    <= '0';
rom_sel    <= '0';

IF  i_lit='0'  THEN
   CASE  i_type  IS
   WHEN op_BRA =>
       CASE i_stack IS
       WHEN op_NONE => IF  condition='1'  THEN
                          sel_ram <= '1';
                       END IF;
       WHEN op_POP  => IF  with_tor='1' AND i_group=op_NZERO  THEN
                          sel_ram <= '1';
                       ELSIF  s_lit='0'  THEN
                          tos_sel <= '1';
                       END IF;
       WHEN op_PUSH =>
          CASE i_group IS
          WHEN op_EXC | op_INT =>
                            write <= '1';
                            sel_ram <= '1';
                            IF  with_tor='1'  THEN
                               tor_sel <= '1';
                            ELSE
                               pc_sel <= '1';
                            END IF;
          WHEN op_IRET   => sel_ram <= '1';
          WHEN op_QOVL   => IF  stBus.flags(s_ovl_bit)='1'  THEN
                               write <= '1';
                               sel_ram <= '1';
                               IF  with_tor='1'  THEN
                                  tor_sel <= '1';
                               ELSE
                                  pc_sel <= '1';
                               END IF;
                            END IF;
          WHEN op_THREAD => IF  with_ip='1'  THEN
                               sel_ram <= '1';
                            END IF;
          WHEN op_TOKEN  => IF  with_ip='1'  THEN
                               sel_ram <= '1';
                               IF  with_tokens='1' AND token='0'  THEN
                                  write <= '1';
                                  IF  with_tor='1'  THEN
                                     tor_sel <= '1';
                                  ELSE
                                     tos_sel <= '1';
                                  END IF;
                               END IF;
                            END IF;
          WHEN OTHERS => NULL;
          END CASE;
       WHEN OTHERS  =>
          IF  with_tor='1'  THEN
             tor_sel <= '1';
          ELSE
             pc_sel <= '1';
          END IF;
          IF  condition='1'  THEN
             write <= '1';
             sel_ram <= '1';
          END IF;
       END CASE;
   WHEN op_ALU => IF  with_tor='1' AND i_stack=op_NONE AND i_group=op_0DIVS  THEN
                     tor_sel <= '1';
                     write <= '1';
                     sel_ram <= '1';
                  END IF;
   WHEN op_MEM =>
       CASE i_stack IS
       WHEN op_NONE =>
           tos_sel <= '1';
           CASE  i_group  IS
           WHEN op_RSTACK => write <= '1';
                             sel_ram <= '1';
                             IF  with_tor='1'  THEN
                                tos_sel <= '0';
                                tor_sel <= '1';
                             END IF;
           WHEN op_TOR   => IF  with_tor='0'  THEN
                               write <= '1';
                               sel_ram <= '1';
                            END IF;
           WHEN op_LOCAL => IF  with_locals='1'  THEN
                               tos_sel <= '0';
                               nos_sel <= '1';
                               write <= '1';
                               sel_ram <= '1';
                            END IF;
           WHEN op_TASK  => IF  with_tasks='1'  THEN
                               nos_sel <= '1';
                               tos_sel <= '0';
                               write <= '1';
                               sel_ram <= '1';
                            END IF;
           WHEN OTHERS   => NULL;
           END CASE;
       WHEN op_POP  =>
           nos_sel <= '1';
           write <= '1';
           IF  regspace='0'  THEN
              IF  iospace='1'  THEN
                 IF  romspace='1'  THEN
                    IF  rom_sel_d='0' AND prog_ram_width/=0  THEN
                       rom_sel <= '1';
                    END IF;
                 ELSE
                    sel_ext <= '1';
                 END IF;
              ELSE
                 sel_ram <= '1';
              END IF;
           ELSE
              internal <= '1';
           END IF;
       WHEN op_PUSH =>
           IF  regspace='0'  THEN
              IF  iospace='1'  THEN
                 IF  romspace='1'  THEN
                    IF  rom_sel_d='0' AND prog_ram_width/=0  THEN
                       rom_sel <= '1';
                    END IF;
                 ELSE
                    sel_ext <= '1';
                 END IF;
              ELSE
                 sel_ram <= '1';
              END IF;
           ELSE
              internal <= '1';
           END IF;
       WHEN OTHERS  => -- op_BOTH
           CASE i_group IS
           WHEN op_STATUS => status_sel <= '1';
           WHEN op_TOR    => IF  with_tor='1'  THEN
                                tor_sel <= '1';
                             ELSE
                                sel_ram <= '1';
                             END IF;
           WHEN op_RSTACK => sel_ram <= '1';
           WHEN op_LOCAL  => IF  with_locals='1'  THEN
                                sel_ram <= '1';
                             END IF;
           WHEN op_RSP    => rsp_sel <= '1';
           WHEN op_DSP    => dsp_sel <= '1';
           WHEN op_TASK   => IF  with_tasks='1'  THEN
                                sel_ram <= '1';
                             END IF;
           WHEN op_IP     => ip_sel <= '1';
           WHEN OTHERS    => NULL;
           END CASE;
       END CASE;
   WHEN OTHERS => -- op_USR
       write <= '1';
       sel_ram <= '1';
       IF  with_tor='1'  THEN
          tor_sel <= '1';
       ELSE
          pc_sel <= '1';
       END IF;   
   END CASE;
END IF;

END PROCESS uBus_control;

rom_sel_inhibit: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      rom_sel_d <= '0';
   ELSIF  rising_edge(clk)  THEN
      IF  clk_en='1' AND prog_ram_width/=0  THEN
         rom_sel_d <= rom_sel;
      END IF;
   END IF;
END PROCESS rom_sel_inhibit;

-- ************** internal memory mapped registers ***************************
-- reg_addr_width address bits are set aside for internal registers at the top
-- of the data memory space.
-- ***************************************************************************

task_sel <= '1'  WHEN  uBus.addr=TASK_REGISTER AND internal='1' AND with_tasks='1'  ELSE '0';

taskreg: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      task_i <= (OTHERS => '0');
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1' AND write='1' AND task_sel='1'  THEN
        task_i <= uBus.data(data_addr_width-1 DOWNTO 0);
     END IF;
   END IF;
END PROCESS taskreg;

END tri_state;
